#include "tpinterface.h"
#include "support.h"
#include "utils.h"
#include "interface.h"
#include <gtk/gtk.h>
#include <pthread.h>

void show_message(const gchar*, const gchar*, gboolean);

gdouble	Progress_X;
INT32	tcpbytes;

const gchar* initialize(gboolean child)
{
    const gchar* tpengine_errmsg = _("Failed to initialize TPTEST engine");
    const gchar* port_errmsg = _("Error: invalid server control port argument");
    const gchar* host_errmsg = _("Error: hostname lookup failed");

    engp = CreateContext();
    if (!engp)
	return tpengine_errmsg;

    sprintf(engp->clientInfo,"%s - GNU/Linux", PACKAGE_STRING);
    strcpy(engp->hostName, TESTSERVER);
    engp->hostCtrlPort = TESTSERVERPORT;
    if (engp->hostCtrlPort == 0) {
	/* error - invalid server port argument */
	return port_errmsg;
    }

    /* check server address argument */
    if (inet_addr(engp->hostName) == (unsigned long) INADDR_NONE)
    {
	struct hostent * hent;
	hent = gethostbyname(engp->hostName);
	if (hent == NULL)
	    return host_errmsg;
	else
	    engp->hostIP.s_addr = ((struct in_addr *)(hent->h_addr))->s_addr;
    }
    else 
	engp->hostIP.s_addr = inet_addr(engp->hostName);

    engp->tpMode = CLM_NONE;
    engp->stats.MajorVersion = MAJORVERSION;
    engp->stats.MinorVersion = MINORVERSION;
    update_testing(engp, child);
    return NULL;
}

void show_message(const gchar* head,
	const gchar* body, gboolean error)
{
    GtkWidget*	dialog;
    int		MSG_TYPE;

    MSG_TYPE = (error)? GTK_MESSAGE_ERROR : GTK_MESSAGE_INFO;
    gdk_threads_enter();
    dialog = messagebox(head, body, MSG_TYPE, GTK_BUTTONS_OK);
    
    gtk_dialog_run((GtkDialog *) dialog);
    gtk_object_destroy((GtkObject *) dialog);
    gdk_threads_leave();
    flush_display();
}

void clear_test_results(gboolean child)
{
    GtkWidget*	TCP = NULL;
    GtkWidget*	UDP = NULL;
    GtkWidget*	TCP_send_best = NULL;
    GtkWidget*	UDP_send_best = NULL;
    GtkWidget*	TCP_receive_best = NULL;
    GtkWidget*	UDP_receive_best = NULL;
    GtkWidget*	best_load_send = NULL;
    GtkWidget*	best_load_receive = NULL;

    if (child)
	gdk_threads_enter();
    TCP = lookup_widget(mainwindow, "TCP_action");
    UDP = lookup_widget(mainwindow, "UDP_action");
    TCP_send_best = lookup_widget(mainwindow, "TCP_send_result");
    UDP_send_best = lookup_widget(mainwindow, "UDP_send_result");
    TCP_receive_best = lookup_widget(mainwindow, "TCP_receive_result");
    UDP_receive_best = lookup_widget(mainwindow, "UDP_receive_result");
    best_load_send = lookup_widget(mainwindow, "total_send");
    best_load_receive = lookup_widget(mainwindow, "total_receive");

    gtk_entry_set_text((GtkEntry *) TCP, "");
    gtk_entry_set_text((GtkEntry *) UDP, "");
    gtk_entry_set_text((GtkEntry *) TCP_send_best, "");
    gtk_entry_set_text((GtkEntry *) TCP_receive_best, "");
    gtk_entry_set_text((GtkEntry *) UDP_send_best, "");
    gtk_entry_set_text((GtkEntry *) UDP_receive_best, "");
    gtk_entry_set_text((GtkEntry *) best_load_send, "");
    gtk_entry_set_text((GtkEntry *) best_load_receive, "");	

    flush_display();
    if (child)
	gdk_threads_leave();
}

void update_testing(TPEngine* engp, gboolean child)
{
    GtkWidget*	TCP = NULL;
    GtkWidget*	UDP = NULL;
    GtkWidget*	TCPinfo = NULL;
    GtkWidget*	UDPinfo = NULL;

    if (child)
	gdk_threads_enter();
    TCP = lookup_widget(mainwindow, "TCP_action");
    UDP = lookup_widget(mainwindow, "UDP_action");
    TCPinfo = lookup_widget(mainwindow, "TCP_info");
    UDPinfo = lookup_widget(mainwindow, "UDP_info");

    gtk_entry_set_text((GtkEntry *) TCP, "");
    gtk_entry_set_text((GtkEntry *) UDP, "");
    gtk_label_set_text((GtkLabel *) TCPinfo, "");
    gtk_label_set_text((GtkLabel *) UDPinfo, "");

    if (my_state != MYSTATE_IDLE)
    {
	tcpbytes = engp->tcpBytes;
	switch (engp->tpMode)
	{
	    case M_TCP_SEND:
		gtk_label_set_text((GtkLabel *) TCPinfo, _("Sending"));
		gtk_entry_set_alignment((GtkEntry *)TCP, 0.5);
		gtk_entry_set_text((GtkEntry *) TCP, UInt32ToString(engp->tcpBytes));
		break;
	    case M_TCP_RECV:
		gtk_label_set_text((GtkLabel *) TCPinfo, _("Receiving"));
		gtk_entry_set_alignment((GtkEntry *)TCP, 0.5);
		gtk_entry_set_text((GtkEntry *) TCP, UInt32ToString(engp->tcpBytes));
		break;
	    case M_UDP_SEND:
		gtk_label_set_text((GtkLabel *) UDPinfo, _("Test sending with"));
		gtk_entry_set_alignment((GtkEntry *)UDP, 0.5);
		gtk_entry_set_text((GtkEntry *) UDP,
			Int32ToString((int)(engp->bitsPerSecond)));
		break;
	    case M_UDP_RECV:
		gtk_label_set_text((GtkLabel *) UDPinfo, _("Test receiving with"));
		gtk_entry_set_alignment((GtkEntry *)UDP, 0.5);
		gtk_entry_set_text((GtkEntry *) UDP,
		    Int32ToString((int)(engp->bitsPerSecond)));
		break;
	}
    }
    flush_display();
    if (child)
	gdk_threads_leave();
}

void update_load_results(TPEngine* engp)
{
    GtkWidget*	best_load_send = NULL;
    GtkWidget*	best_load_receive = NULL;
    gdouble	sendres = 0.0, recvres = 0.0;
    gchar*	tmp = NULL;

    gdk_threads_enter();
    best_load_send = lookup_widget(mainwindow, "total_send");
    best_load_receive = lookup_widget(mainwindow, "total_receive");

    if (engp->bestTCPSendRate > 0.0 && engp->bestUDPSendRate > 0.0)
    {
	sendres = (((double)engp->bestTCPSendRate / 
		    (double)engp->bestUDPSendRate) * 100.0);
	if (sendres > 100.0)
	    sendres = 100.0;
	tmp = g_strdup_printf("%02.1f %%", sendres);
	gtk_entry_set_alignment((GtkEntry *) best_load_send, 0.5);
	gtk_entry_set_text((GtkEntry *) best_load_send, tmp);
	g_free(tmp);
    }
    if (engp->bestTCPRecvRate > 0.0 && engp->bestUDPRecvRate > 0.0)
    {
	recvres = (((double)engp->bestTCPRecvRate /
		    (double)engp->bestUDPRecvRate) * 100.0);
	if (recvres > 100.0) recvres = 100.0;
	tmp = g_strdup_printf("%02.1f %%", recvres);
	gtk_entry_set_alignment((GtkEntry *) best_load_receive, 0.5);
	gtk_entry_set_text((GtkEntry *) best_load_receive, tmp);
	g_free(tmp);
    }
    flush_display();
    gdk_threads_leave();
}

void update_best_rate(TPEngine* engp)
{
    GtkWidget*	TCP_send_best = NULL;
    GtkWidget*	UDP_send_best = NULL;
    GtkWidget*	TCP_receive_best = NULL;
    GtkWidget*	UDP_receive_best = NULL;

    gdk_threads_enter();
    TCP_send_best = lookup_widget(mainwindow, "TCP_send_result");
    UDP_send_best = lookup_widget(mainwindow, "UDP_send_result");
    TCP_receive_best = lookup_widget(mainwindow, "TCP_receive_result");
    UDP_receive_best = lookup_widget(mainwindow, "UDP_receive_result");

    if (engp->bestTCPSendRate == 0.0)
	gtk_entry_set_text((GtkEntry *) TCP_send_best, "");
    else
    {
	gtk_entry_set_alignment((GtkEntry *)TCP_send_best, 0.5);
	gtk_entry_set_text((GtkEntry *) TCP_send_best,
		Int32ToString((int)(engp->bestTCPSendRate * 8.0)));
    }

    if (engp->bestTCPRecvRate == 0.0)
	gtk_entry_set_text((GtkEntry *) TCP_receive_best, "");
    else
    {
	gtk_entry_set_alignment((GtkEntry *)TCP_receive_best, 0.5);
	gtk_entry_set_text((GtkEntry *) TCP_receive_best,
		Int32ToString((int)(engp->bestTCPRecvRate * 8.0)));
    }

    if (engp->bestUDPSendRate == 0.0)
	gtk_entry_set_text((GtkEntry *) UDP_send_best, "");
    else
    {
	gtk_entry_set_alignment((GtkEntry *)UDP_send_best, 0.5);
	gtk_entry_set_text((GtkEntry *) UDP_send_best,
		Int32ToString((int)(engp->bestUDPSendRate * 8.0)));
    }

    if (engp->bestUDPRecvRate == 0.0)
	gtk_entry_set_text((GtkEntry *) UDP_receive_best, "");
    else
    {
	gtk_entry_set_alignment((GtkEntry *)UDP_receive_best, 0.5);
	gtk_entry_set_text((GtkEntry *) UDP_receive_best,
		Int32ToString((int)(engp->bestUDPRecvRate * 8.0)));
    }
    flush_display();
    gdk_threads_leave();
}

void* tptest(void* dummy)
{
    GtkButton* button = NULL;
    struct timespec sleeptime;
    gboolean inner_error;

    /* 0.5 sec wait between automatic tests */
    sleeptime.tv_sec = 0;
    sleeptime.tv_nsec = 500000000;

    sched_yield();
    while (TRUE)
    {
	switch (my_state)
	{
	    case MYSTATE_IDLE:
		gdk_threads_enter();
		button = (GtkButton *) lookup_widget(mainwindow, "button1");
		if (button) {
		    gtk_button_set_label(button, _("_START"));
		    flush_display();
		}
		gdk_threads_leave();
		return (void *) NULL;
		break;
	    case MYSTATE_STDTEST:
		time(&starttime);
		inner_error = FALSE;
		while(my_state != MYSTATE_IDLE)
		{
		    engp->tpMode = AdvanceTest(engp, CLM_AUTO, engp->tpMode, 0);
		    if (engp->tpMode == CLM_NONE)
		    {
			time(&stoptime);
			my_state = MYSTATE_IDLE;
			update_testing(engp, TRUE);
			update_load_results(engp);
			init_progressbar();
			show_message(
				_("Testing"),
				_("Test finished succesfully"),
				FALSE);
			break;
		    }
		    if (StartClientContext(engp) != 0)
		    {
		        my_state = MYSTATE_IDLE;
		        do_error(_("StartClientContext failed"));
		        break;
		    }
		    if (inner_error)
		    {
			my_state = MYSTATE_IDLE;
			StopContext(engp);
			init_progressbar();
			ClearBestRates(engp);
			clear_test_results(FALSE);
			break;
		    }
		    update_testing(engp, TRUE);
		    update_best_rate(engp);
		    Progress_X = 0;
		    init_progressbar();
		    while(my_state != MYSTATE_IDLE)
		    {
			RunClientContext(engp);
			update_progressbar(engp);
			if (engp->state == CLSM_FAILED)
		        {
			    my_state = MYSTATE_IDLE;
			    do_error(_("Test failed"));
			    inner_error = TRUE;
			    break;
			}
			else if (engp->state == CLSM_COMPLETE)
			{
			    fill_progressbar();
			    update_testing(engp, TRUE);
			    break;
			}
			sched_yield();
		    }
		    /* sleep 0.5 seconds before starting next test, if any */
		    nanosleep(&sleeptime, NULL);
		    sched_yield();
		}
		time(&stoptime);
		my_state = MYSTATE_IDLE;
		break;
	}
	sched_yield();
    }
    return (void *) NULL;
}

void init_progressbar(void)
{
    GtkProgressBar  *progress = NULL;
    gdouble	    current;
    gchar*	    text = NULL;

    gdk_threads_enter();
    progress = (GtkProgressBar*) lookup_widget(mainwindow, "progressbar1");
    current = 0.0;
    text = g_strdup_printf(_("%3.0f%% completed"), current * 100);
    g_object_set(progress, "fraction", current, "text", text, NULL);
    g_free(text);
    flush_display();
    gdk_threads_leave();
}

void fill_progressbar(void)
{
    GtkProgressBar  *progress = NULL;
    gdouble	    current;
    gchar*	    text = NULL;

    gdk_threads_enter();
    progress = (GtkProgressBar*) lookup_widget(mainwindow, "progressbar1");
    current = 1.0;
    text = g_strdup_printf(_("%3.0f%% completed"), current * 100);
    g_object_set(progress, "fraction", current, "text", text, NULL);
    g_free(text);
    flush_display();
    gdk_threads_leave();
}

void update_progressbar(TPEngine* engp)
{
    GtkProgressBar  *progress = NULL;
    gdouble	    current = 0.0;
    gdouble	    new_x = 0.0;
    gchar*	    text = NULL;
   
    gdk_threads_enter();
    switch (engp->tpMode)
    {
	case M_UDP_SEND:
	    if (engp->nPackets > 0)
		new_x = ((gdouble) engp->curSend) / 
			((gdouble) engp->nPackets);
	    break;
	case M_UDP_RECV:
	    if (engp->nPackets > 0)
		new_x = ((gdouble) engp->stats.PktsRecvd) / 
			((gdouble) engp->nPackets);
	    break;
	case M_TCP_SEND:
	    if (engp->tcpBytes > 0)
		new_x = ((gdouble) engp->stats.BytesSent) / 
			((gdouble) engp->tcpBytes);
	    break;
	case M_TCP_RECV:
	    if (engp->tcpBytes > 0)
		new_x = ((gdouble) engp->stats.BytesRecvd) /
			((gdouble) engp->tcpBytes);
	    break;
    }
    
    if (new_x > 1)
	new_x = 1;

    if (new_x > Progress_X)
    {
	progress = (GtkProgressBar*) lookup_widget(mainwindow, "progressbar1");
	g_object_get(progress, "fraction", &current, NULL);
	current = Progress_X = new_x;
	current = CLAMP(current, 0.0, 0.999);
	text = g_strdup_printf(_("%3.0f%% completed"), current * 100);
	g_object_set(progress, "fraction", current, "text", text, NULL);
	g_free(text);
	flush_display();
    }
    gdk_threads_leave();
}

void do_error(gchar *str)
{
    gchar *str1 = NULL;

    str1 = g_strdup_printf("Error %d", (int) engp->failCode);
    switch (engp->failCode) {
    case TPER_TIMEOUT:
	str = g_strconcat(str, _("- test timed out"), NULL);
	break;
    case TPER_SERVERBUSY:
	str = g_strconcat(str, _("- server busy. Try again later"), NULL);
	break;
    case TPER_SERVERDENY:
	str = g_strconcat(str, _("- server refused us. Try another server"), NULL);
	break;
    case TPER_CONNECTFAIL:
	str = g_strconcat(str, _("- server is not running. Try another server"), NULL);
	break;
    case TPER_MASTERBUSY:
	str = g_strconcat(str, _("- master server busy. Try again later"), NULL);
	break;
    case TPER_MASTERDENIED:
	str = g_strconcat(str, _("- master server refused us"), NULL);
	break;
    case TPER_USERABORT:
	str = g_strconcat(str, _("- user abort"), NULL);
	break;
    case TPER_NATFAIL:
	str = g_strconcat(str, _("- NATOPEN failed. Do you have a firewall?"), NULL);
	break;
    case TPER_SRVABORT:
	str = g_strconcat(str, _("- server aborted test"), NULL);
	break;
    case TPER_CTRLCLOSED:
	str = g_strconcat(str, _("- server closed control connection"), NULL);
	break;
    default:
	str = g_strconcat(str, _(" (last msg: "), NULL);
	str = g_strconcat(str, engp->ctrlMessage, NULL);
	str = g_strconcat(str, ")", NULL);
	break;
    }
    my_state = MYSTATE_IDLE;
    update_testing(engp, TRUE);
    str = g_strdup_printf("%s: %s\n", str1, str);
    show_message(
	    _("Testing"),
	    str,
	    TRUE);
    g_free(str1);
}
